from functools import wraps

import numpy as np

from stats_arrays.errors import MultipleRowParamsArrayError

BASE_DTYPE = [
    ("loc", np.float64),
    ("scale", np.float64),
    ("shape", np.float64),
    ("minimum", np.float64),
    ("maximum", np.float64),
    ("negative", bool),
]


def flatten_numpy_array(obj):
    if not isinstance(obj, np.ndarray):
        return obj
    return obj.ravel()


def one_row_params_array(function):
    @wraps(function)
    def wrapper(cls, params, *args, **kwargs):
        if len(params.shape) == 1:
            params = params.reshape(params.shape[0], 1)
        else:
            if params.shape[0] != 1:
                raise MultipleRowParamsArrayError
        # Flatten any additional inputs to one dimension
        # Needed for PDF optional xs input
        args = [flatten_numpy_array(x) for x in args]
        kwargs = {key: flatten_numpy_array(obj) for key, obj in kwargs.items()}
        return function(cls, params, *args, **kwargs)

    return wrapper


def construct_params_array(length=1, include_type=False):
    dtype = BASE_DTYPE
    if include_type:
        dtype = dtype + [("uncertainty_type", np.uint8)]
    params = np.zeros((length,), dtype=dtype)
    params["minimum"] = params["maximum"] = np.nan
    params["scale"] = params["loc"] = params["shape"] = np.nan
    return params



from __future__ import division

import numpy as np
from scipy import stats

from stats_arrays.errors import InvalidParamsError
from stats_arrays.utils import one_row_params_array
from stats_arrays.distributions.base import UncertaintyBase


class LognormalUncertainty(UncertaintyBase):
    id = 2
    description = "Lognormal uncertainty"

    @classmethod
    def validate(cls, params):
        """Custom validation because mean gets log-transformed"""
        if np.isnan(params["loc"]).sum():
            raise InvalidParamsError(
                "Real location (mu) values are required for" " lognormal uncertainties."
            )
        if np.isnan(params["scale"]).sum() or (params["scale"] <= 0).sum():
            raise InvalidParamsError(
                "Real, positive scale (sigma) values are required for"
                " lognormal uncertainties."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        data = seeded_random.lognormal(
            params["loc"], params["scale"], size=(size, params.shape[0])  # Mu  # Sigma
        ).T
        # Negative is needed because log loses sign information.
        # Error handling not included, as this loop is called many times
        data[params["negative"], :] = -1 * data[params["negative"], :]
        return data

    @classmethod
    def cdf(cls, params, vector):
        vector[params["negative"]] = -1 * vector[params["negative"]]
        if vector.ndim == 1 or vector.shape[1] == 1:
            # Can do all at once, **much** faster
            results = stats.lognorm.cdf(
                vector.reshape((-1,)), params["scale"], scale=np.exp(params["loc"])
            ).reshape((-1, 1))
        else:
            vector = cls.check_2d_inputs(params, vector)
            results = np.zeros(vector.shape)
            for row in range(params.shape[0]):
                results[row, :] = stats.lognorm.cdf(
                    vector[row, :],
                    params["scale"][row],
                    scale=np.exp(params["loc"][row]),
                )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        if percentages.ndim == 1 or percentages.shape[1] == 1:
            # Can do all at once, **much** faster
            results = stats.lognorm.ppf(
                percentages.reshape((-1,)), params["scale"], scale=np.exp(params["loc"])
            ).reshape((-1, 1))
        else:
            percentages = cls.check_2d_inputs(params, percentages)
            results = np.zeros(percentages.shape)
            for row in range(percentages.shape[0]):
                results[row, :] = stats.lognorm.ppf(
                    percentages[row, :],
                    params["scale"][row],
                    scale=np.exp(params["loc"][row]),
                )
        results[params["negative"]] = -1 * results[params["negative"]]
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        negative = -1 if bool(params["negative"]) else 1
        geometric_mu = float(np.exp(params["loc"]))
        sigma = float(params["scale"])
        mu = float(params["loc"])
        geometric_sigma = float(np.exp(sigma))
        mean = np.exp(mu + (sigma**2) / 2)
        mode = np.exp(mu - sigma**2)
        ci_95_lower = geometric_mu / (geometric_sigma**2)
        ci_95_upper = geometric_mu * (geometric_sigma**2)
        if negative == -1:
            ci_95_lower, ci_95_upper = ci_95_upper, ci_95_lower
        return {
            "median": negative * geometric_mu,
            "mode": negative * mode,
            "mean": negative * mean,
            "lower": negative * ci_95_lower,
            "upper": negative * ci_95_upper,
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        """Generate probability distribution function for lognormal distribution."""
        n = params["negative"]
        if xs is None:
            # Find nice range to graph
            lower = (
                (-1 if n else 1)
                * params["loc"]
                / (np.exp(params["scale"]) ** cls.standard_deviations_in_default_range)
            )[0, 0]
            upper = (
                (-1 if n else 1)
                * params["loc"]
                * (np.exp(params["scale"]) ** cls.standard_deviations_in_default_range)
            )[0, 0]
            if n:
                lower, upper = upper, lower

            xs = np.linspace(
                (lower if np.isnan(params["minimum"]) else params["minimum"]).ravel(),
                (upper if np.isnan(params["maximum"]) else params["maximum"]).ravel(),
                cls.default_number_points_in_pdf,
            ).ravel()

        ys = stats.lognorm.pdf(
            -1 * xs if n else xs, params["scale"], scale=np.exp(params["loc"])
        )
        return xs, ys.ravel()



from __future__ import division

import numpy as np
from scipy import stats

from stats_arrays.errors import ImproperBoundsError
from stats_arrays.utils import one_row_params_array
from stats_arrays.distributions.base import BoundedUncertaintyBase


class UniformUncertainty(BoundedUncertaintyBase):
    """Continuous uniform distribution. In SciPy, the uniform distribution is defined from loc to loc+scale."""

    id = 4
    description = "Uniform uncertainty"

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        return seeded_random.uniform(
            params["minimum"],  # Minimum (low)
            params["maximum"],  # Maximum (high)
            size=(size, params.shape[0]),
        ).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = np.zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.uniform.cdf(
                vector[row, :],
                loc=params[row]["minimum"],
                scale=params[row]["maximum"] - params[row]["minimum"],
            )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        scale = (params["maximum"] - params["minimum"]).reshape(params.shape[0], 1)
        return percentages * scale + params["minimum"].reshape(params.shape[0], 1)

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        mean = (params["maximum"] + params["minimum"]) / 2
        return {
            "mean": mean,
            "mode": mean,
            "median": mean,
            "lower": params["minimum"],
            "upper": params["maximum"],
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs is None:
            xs = (params["minimum"], params["maximum"])
        percentage = float((1 / (params["maximum"] - params["minimum"])).flat[0])
        ys = np.array([percentage for x in xs])
        return np.array([float(x.flat[0]) for x in xs]), ys


class TriangularUncertainty(BoundedUncertaintyBase):
    id = 5
    description = "Triangular uncertainty"

    @classmethod
    def validate(cls, params):
        super(TriangularUncertainty, cls).validate(params)
        if (params["loc"] > params["maximum"]).sum() or (
            params["loc"] < params["minimum"]
        ).sum():
            raise ImproperBoundsError("Most likely value outside the given bounds.")

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        return seeded_random.triangular(
            params["minimum"],  # Left
            params["loc"],  # Mode
            params["maximum"],  # Right
            size=(size, params.shape[0]),
        ).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        # Adjust parameters to (0,1) range
        adjusted_means, scale = cls.rescale(params)
        # To be broadcasted correctly, scale and mins must be column vectors
        scale.resize(scale.shape[0], 1)
        mins = np.array(params["minimum"])
        mins.resize(params.shape[0], 1)
        # Adjust values to use cdf for to (0,1) range
        adjusted_vector = (vector - mins) / scale
        results = np.zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.triang.cdf(
                adjusted_vector[row, :], adjusted_means[row]
            )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        adjusted_means, scale = cls.rescale(params)
        scale.resize(scale.shape[0], 1)
        adjusted_means.resize(scale.shape[0], 1)
        mins = np.array(params["minimum"])
        mins.resize(params.shape[0], 1)
        return stats.triang.ppf(percentages, adjusted_means) * scale + mins

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        mode = float(params["loc"].flat[0])
        mean = float(((params["minimum"] + params["maximum"] + mode) / 3).flat[0])
        lower, median, upper = cls.ppf(
            params, np.array([[0.0125, 0.5, 0.9875]])
        ).ravel()
        return {
            "mean": mean,
            "median": float(median.flat[0]),
            "mode": mode,
            "lower": float(lower.flat[0]),
            "upper": float(upper.flat[0]),
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs is None:
            lower = params["minimum"]
            upper = params["maximum"]
            mode = params["loc"]
            if not mode:
                mode = (upper + lower) / 2
            xs = np.array([float(x.flat[0]) for x in (lower, mode, upper)])
            ys = np.array([0, float(((mode - lower) / (upper - lower)).flat[0]), 0])
        else:
            adjusted_means, scale = cls.rescale(params)
            adj_xs = (xs - params["minimum"]) / scale
            ys = stats.triang.pdf(adj_xs, adjusted_means)
        return xs, ys



from __future__ import division

import numpy as np

from stats_arrays.errors import InvalidParamsError
from stats_arrays.distributions.base import UncertaintyBase


class GammaUncertainty(UncertaintyBase):
    """
    The Gamma uncertainty distribution probability density function as a function of :math:`k`, the shape parameters, and :math:`\\theta`, the scale parameter:

    .. math:: f(x;k,\\theta) =  \\frac{x^{k-1}e^{-\\frac{x}{\\theta}}}{\\theta^k\\Gamma(k)}

    The scale parameter :math:`k` is ``shape``, and :math:`\\theta` is ``scale``. An optional location parameter, which offsets the distribution from the origin, can be specified in ``loc``.

    See https://en.wikipedia.org/wiki/Gamma_distribution.
    """

    id = 9
    description = "Gamma uncertainty"

    @classmethod
    def validate(cls, params, transform=False):
        if (params["shape"] <= 0).sum() or np.isnan(params["shape"]).sum():
            raise InvalidParamsError(
                "Positive shape (k) values required for Gamma distribution."
            )
        if (params["scale"] <= 0).sum() or np.isnan(params["scale"]).sum():
            raise InvalidParamsError(
                "Positive scale (theta) values required for Gamma distribution."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None, **kwargs):
        if seeded_random is None:
            seeded_random = np.random
        offset = params["loc"].copy()
        offset[np.isnan(offset)] = 0
        data = (
            offset.reshape((-1, 1))
            + seeded_random.gamma(
                shape=params["shape"],
                scale=params["scale"],
                size=(size, params.shape[0]),
            ).T
        )
        data[params["negative"], :] = -1 * data[params["negative"], :]
        return data



from __future__ import division

import numpy as np

from stats_arrays.errors import InvalidParamsError
from stats_arrays.distributions.base import UncertaintyBase


class WeibullUncertainty(UncertaintyBase):
    """
    The Weibull distribution has the probability distribution function:

    .. math:: f(x; k, \\lambda) = \\frac{k}{\\lambda} \\left( \\frac{x}{\\lambda} \\right)^{k - 1} e^{- \\left( x / \\lambda \\right)^{k}}

    In our implementation, :math:`\\lambda` is ``scale``, and :math:`k`  is ``shape``. An optional location parameter, which offsets the distribution from the origin, can be specified in ``loc``.

    See https://en.wikipedia.org/wiki/Weibull_distribution.
    """

    id = 8
    description = "Weibull uncertainty"

    @classmethod
    def validate(cls, params):
        if (params["shape"] <= 0).sum():
            raise InvalidParamsError("Real, positive ``k`` values need for Weibull.")
        if (params["scale"] <= 0).sum():
            raise InvalidParamsError(
                "Real, positive ``lambda`` values need for Weibull."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None, transform=False):
        if seeded_random is None:
            seeded_random = np.random
        offset = params["loc"].copy()
        offset[np.isnan(offset)] = 0
        data = (
            offset.reshape((-1, 1))
            + params["scale"].reshape((-1, 1))
            * seeded_random.weibull(params["shape"], size=(size, params.shape[0])).T
        )
        data[params["negative"], :] = -1 * data[params["negative"], :]
        return data



class StatsArraysError(Exception):
    """Base class for `stats_arrays` errors"""


class ImproperBoundsError(StatsArraysError):
    """The bounds and mean specified are illegal for this distribution."""


class MaximumIterationsError(StatsArraysError):
    """Drawing random numbers from the distribution to fit in the bounds specified used more than the maximum number of iterations allowed."""


class UnknownUncertaintyType(StatsArraysError):
    """The uncertainty type is not defined in uncertainty_choices."""


class UndefinedDistributionError(StatsArraysError):
    """Values were attempted to be calculated by an undefined distribution."""


class InvalidParamsError(StatsArraysError):
    """Invalid params array passed to uncertainty distribution init."""


class MultipleRowParamsArrayError(StatsArraysError):
    """A function or method which doesn't accept it was passed a params array with more than one row."""


class UnreasonableBoundsError(StatsArraysError):
    """The provided bounds cover an unreasonably small section of the distribution sample space."""



import numpy as np
import pytest

from stats_arrays.distributions import (
    BernoulliUncertainty,
    LognormalUncertainty,
    NormalUncertainty,
    NoUncertainty,
    TriangularUncertainty,
    UndefinedUncertainty,
    UniformUncertainty,
)
from stats_arrays.random import LatinHypercubeRNG


def make_params_array(length=1):
    assert isinstance(length, int)
    params = np.zeros(
        (length,),
        dtype=[
            ("uncertainty_type", "i2"),
            ("input", "u4"),
            ("output", "u4"),
            ("loc", "f4"),
            ("negative", "b1"),
            ("scale", "f4"),
            ("minimum", "f4"),
            ("maximum", "f4"),
        ],
    )
    params["minimum"] = params["maximum"] = params["scale"] = np.nan
    return params


def test_known_seed():
    params = make_params_array(1)
    params["uncertainty_type"] = UniformUncertainty.id
    params["maximum"] = 1
    params["minimum"] = 0
    params["loc"] = 0.5
    params["scale"] = np.nan
    lhc = LatinHypercubeRNG(params, seed=11111, samples=9)
    assert np.allclose(next(lhc), np.array([[0.5]]))
    assert np.allclose(next(lhc), np.array([[0.7]]))
    assert np.allclose(next(lhc), np.array([[0.5]]))
    assert np.allclose(next(lhc), np.array([[0.6]]))


def test_no_uncertainty_intervals():
    # Two-dimensional array check
    params = make_params_array(2)
    params["uncertainty_type"] = NoUncertainty.id
    params["loc"] = (1, 2)
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        ([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [2, 2, 2, 2, 2, 2, 2, 2, 2, 2])
    )
    assert np.allclose(lhc.hypercube, test_values)
    params["uncertainty_type"] = UndefinedUncertainty.id
    lhc = LatinHypercubeRNG(params)
    assert np.allclose(lhc.hypercube, test_values)
    # One-dimensional array check
    params = make_params_array(1)
    params["loc"] = 1
    params["uncertainty_type"] = NoUncertainty.id
    lhc = LatinHypercubeRNG(params)
    test_values = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


def test_lognormal_intervals_2d():
    # Two-dimensional array check
    params = make_params_array(2)
    params["uncertainty_type"] = LognormalUncertainty.id
    params["loc"] = (1, 2)
    params["scale"] = (1, 2)
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            [
                0.71521093,
                1.09586291,
                1.48499978,
                1.91792577,
                2.42495914,
                3.0470846,
                3.85262859,
                4.97579579,
                6.74268247,
                10.3312957,
            ],
            [
                0.51152672,
                1.20091561,
                2.20522451,
                3.67843954,
                5.88042728,
                9.28472525,
                14.84274814,
                24.75854564,
                45.4637703,
                106.7356788,
            ],
        ]
    )
    assert np.allclose(lhc.hypercube, test_values)


def test_lognormal_intervals_1d():
    # One-dimensional array check
    params = make_params_array(1)
    params["loc"] = 1
    params["scale"] = 1
    params["uncertainty_type"] = LognormalUncertainty.id
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            0.71521093,
            1.09586291,
            1.48499978,
            1.91792577,
            2.42495914,
            3.0470846,
            3.85262859,
            4.97579579,
            6.74268247,
            10.3312957,
        ]
    ).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


@pytest.mark.skip()
def test_lognormal_intervals_lower():
    # Lower bound check
    params = make_params_array(1)
    params["scale"] = 1
    params["uncertainty_type"] = LognormalUncertainty.id
    params["loc"] = 1
    params["minimum"] = 1
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            [
                1.12095981,
                1.25845417,
                1.41730289,
                1.604463,
                1.83049303,
                2.11247151,
                2.48049433,
                2.99457872,
                3.8006714,
                5.42285066,
            ]
        ]
    ).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


@pytest.mark.skip()
def test_lognormal_intervals_upper():
    # Upper bound check
    params = make_params_array(1)
    params["scale"] = 1
    params["uncertainty_type"] = LognormalUncertainty.id
    params["loc"] = 1
    params["maximum"] = 1
    params["minimum"] = np.nan
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            [
                0.18440486,
                0.26311141,
                0.33393679,
                0.40314545,
                0.47337916,
                0.54630091,
                0.62326149,
                0.70556548,
                0.79462568,
                0.89209264,
            ]
        ]
    ).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


@pytest.mark.skip()
def test_lognormal_intervals_both():
    # Both bounds check
    params = make_params_array(1)
    params["scale"] = 1
    params["uncertainty_type"] = LognormalUncertainty.id
    params["loc"] = 1
    params["minimum"] = 0.5
    params["maximum"] = 1.5
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            [
                0.56110473,
                0.62514369,
                0.69285366,
                0.76502388,
                0.84254518,
                0.92646272,
                1.01804071,
                1.11884934,
                1.23088804,
                1.35676875,
            ]
        ]
    ).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


def test_normal_intervals_2d():
    # Two-dimensional array check
    params = make_params_array(2)
    params["uncertainty_type"] = NormalUncertainty.id
    params["loc"] = (1, 2)
    params["scale"] = (1, 2)
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        (
            [
                -0.33517774,
                0.09154213,
                0.39541465,
                0.6512443,
                0.88581471,
                1.11418529,
                1.3487557,
                1.60458535,
                1.90845787,
                2.33517774,
            ],
            [
                -0.67035547,
                0.18308426,
                0.79082931,
                1.30248861,
                1.77162941,
                2.22837059,
                2.69751139,
                3.20917069,
                3.81691574,
                4.67035547,
            ],
        )
    )
    assert np.allclose(lhc.hypercube, test_values)


def test_normal_intervals_1d():
    # One-dimensional array check
    params = make_params_array(1)
    params["loc"] = 1
    params["scale"] = 1
    params["uncertainty_type"] = NormalUncertainty.id
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            -0.33517774,
            0.09154213,
            0.39541465,
            0.6512443,
            0.88581471,
            1.11418529,
            1.3487557,
            1.60458535,
            1.90845787,
            2.33517774,
        ]
    ).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


@pytest.mark.skip()
def test_normal_intervals_lower():
    # Lower bound check
    params = make_params_array(1)
    params["loc"] = 1
    params["scale"] = 1
    params["uncertainty_type"] = NormalUncertainty.id
    params["minimum"] = 1
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            [
                1.11418529,
                1.22988412,
                1.3487557,
                1.47278912,
                1.60458535,
                1.74785859,
                1.90845787,
                2.09680356,
                2.33517774,
                2.69062163,
            ]
        ]
    ).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


@pytest.mark.skip()
def test_normal_intervals_upper():
    # Upper bound check
    params = make_params_array(1)
    params["loc"] = 1
    params["scale"] = 1
    params["uncertainty_type"] = NormalUncertainty.id
    params["maximum"] = 1
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            [
                -0.69062163,
                -0.33517774,
                -0.09680356,
                0.09154213,
                0.25214141,
                0.39541465,
                0.52721088,
                0.6512443,
                0.77011588,
                0.88581471,
            ]
        ]
    ).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


@pytest.mark.skip()
def test_normal_intervals_both():
    # Both bounds check
    params = make_params_array(1)
    params["loc"] = 1
    params["scale"] = 1
    params["uncertainty_type"] = NormalUncertainty.id
    params["minimum"] = 0.5
    params["maximum"] = 1.5
    lhc = LatinHypercubeRNG(params)
    test_values = np.array(
        [
            [
                0.59665956,
                0.68968379,
                0.78009253,
                0.86873532,
                0.95635658,
                1.04364342,
                1.13126468,
                1.21990747,
                1.31031621,
                1.40334044,
            ]
        ]
    ).reshape(1, 10)
    assert np.allclose(lhc.hypercube, test_values)


def test_uniform_intervals():
    # Two-dimensional array check
    params = make_params_array(2)
    params["uncertainty_type"] = UniformUncertainty.id
    params["maximum"] = 2
    params["minimum"] = 1
    params["loc"] = 1.5
    params["scale"] = np.nan
    lhc = LatinHypercubeRNG(params, samples=9)
    assert np.allclose(
        lhc.hypercube,
        np.array(
            [
                [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9],
                [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9],
            ]
        ),
    )

    # One-dimensional array check
    params = make_params_array(1)
    params["uncertainty_type"] = UniformUncertainty.id
    params["maximum"] = 2
    params["minimum"] = 1
    params["loc"] = 1.5
    params["scale"] = np.nan
    lhc = LatinHypercubeRNG(params, samples=9)
    assert np.allclose(
        lhc.hypercube, np.array([[1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9]])
    )


def test_triangular_intervals():
    # Two-dimensional array check
    params = make_params_array(2)
    params["uncertainty_type"] = TriangularUncertainty.id
    params["maximum"] = 2
    params["minimum"] = 1
    params["loc"] = 1.6
    params["scale"] = np.nan
    lhc = LatinHypercubeRNG(params, samples=9)
    assert np.allclose(
        lhc.hypercube,
        np.array(
            [
                [
                    1.24494898,
                    1.34641017,
                    1.42426408,
                    1.48989796,
                    1.54772257,
                    1.60000001,
                    1.65358985,
                    1.7171573,
                    1.80000001,
                ],
                [
                    1.24494898,
                    1.34641017,
                    1.42426408,
                    1.48989796,
                    1.54772257,
                    1.60000001,
                    1.65358985,
                    1.7171573,
                    1.80000001,
                ],
            ]
        ),
    )

    # One-dimensional array check
    params = make_params_array(1)
    params["uncertainty_type"] = TriangularUncertainty.id
    params["maximum"] = 2
    params["minimum"] = 1
    params["loc"] = 1.6
    params["scale"] = np.nan
    lhc = LatinHypercubeRNG(params, samples=9)
    assert np.allclose(
        lhc.hypercube,
        np.array(
            [
                [
                    1.24494898,
                    1.34641017,
                    1.42426408,
                    1.48989796,
                    1.54772257,
                    1.60000001,
                    1.65358985,
                    1.7171573,
                    1.80000001,
                ]
            ]
        ),
    )


def test_bernoulli():
    # Two-dimensional array check
    params = make_params_array(2)
    params["uncertainty_type"] = BernoulliUncertainty.id
    params["maximum"] = 2
    params["minimum"] = 1
    params["loc"] = 1.6
    lhc = LatinHypercubeRNG(params)
    assert np.allclose(
        lhc.hypercube,
        np.array([[1, 1, 1, 1, 1, 1, 2, 2, 2, 2], [1, 1, 1, 1, 1, 1, 2, 2, 2, 2]]),
    )

    # One-dimensional array check
    params = make_params_array(1)
    params["uncertainty_type"] = BernoulliUncertainty.id
    params["maximum"] = 2
    params["minimum"] = 1
    params["loc"] = 1.6
    lhc = LatinHypercubeRNG(params)
    assert np.allclose(lhc.hypercube, np.array([1, 1, 1, 1, 1, 1, 2, 2, 2, 2]))


def test_heterogeneous_params():
    params = make_params_array(2)
    params["uncertainty_type"][0] = TriangularUncertainty.id
    params["uncertainty_type"][1] = UniformUncertainty.id
    params["maximum"] = 2
    params["minimum"] = 1
    params["loc"] = 1.6
    params["scale"] = np.nan
    lhc = LatinHypercubeRNG(params, samples=9)
    assert np.allclose(
        lhc.hypercube,
        np.array(
            [
                [
                    1.24494898,
                    1.34641017,
                    1.42426408,
                    1.48989796,
                    1.54772257,
                    1.60000001,
                    1.65358985,
                    1.7171573,
                    1.80000001,
                ],
                [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9],
            ]
        ),
    )



import numpy as np

from stats_arrays.errors import UnknownUncertaintyType
from stats_arrays.uncertainty_choices import uncertainty_choices

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable


class RandomNumberGenerator(Iterable):

    def __init__(
        self,
        uncertainty_type,
        params,
        size=1,
        maximum_iterations=100,
        seed=None,
        **kwargs,
    ):
        """
        Create a random number generator from a :ref:`params-array` and an uncertainty distribution.

        Upon instantiation, the class checks that:

        * The ``minimum`` and ``maximum`` bounds, if any, are reasonable
        * The given uncertainty type can be used

        ``uncertainty_type`` is not required to be a subclass of :ref:`UncertaintyBase`, but needs to have the method ``bounded_random_variables``.

        The returned class instance can be called directly::

            >>> from stats_arrays import RandomNumberGenerator, TriangularUncertainty
            >>> params = TriangularUncertainty.from_dicts(
            ...     {'loc': 5, 'minimum': 3, 'maximum': 10},
            ...     {'loc': 1, 'minimum': 0.7, 'maximum': 4.4}
            ...     )
            >>> rng = RandomNumberGenerator(TriangularUncertainty, params)
            >>> rng.generate_random_numbers()
            array([[ 8.00843856],
               [ 1.54968237]])

        but can also be used as an iterator::

            >>> zip(range(2), rng)
            [(0, array([[ 5.34298156],
               [ 1.02447677]])),
             (1, array([[ 5.45360508],
               [ 1.99372889]]))]

        Args:
            * **uncertainty_type** (object): An uncertainty type class (subclass of ``stats_arrays.distributions.UncertaintyBase``)
            * **params** (array): The :ref:`params-array`
            * *size* (int, optional): The number of samples to draw from each parameter. Default is ``1``.
            * *maximum_iterations* (int, optional): The number of times to draw samples that fit within the given bounds, if any, before raising ``stats_arrays.MaximumIterationsError``. Default is ``100``.
            * *seed* (int, optional): Seed value for the random number generator. Default is ``None``.

        Returns:
            A class instance

        """
        self.params = params
        self.length = self.params.shape[0]
        self.size = size
        self.uncertainty_type = uncertainty_type
        self.maximum_iterations = maximum_iterations
        self.random = np.random.RandomState(seed)
        self.verify_uncertainty_type()
        self.verify_params()

    def verify_params(self, params=None, uncertainty_type=None):
        """Verify that parameters are within bounds. Mean is not restricted to bounds, unless the distribution requires it (e.g. triangular)."""
        if params is None:  # Can't convert array to boolean
            params = self.params
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        uncertainty_type.validate(params)

    def verify_uncertainty_type(self, uncertainty_type=None):
        """Make sure the given uncertainty type provides the method ``bounded_random_variables``."""
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        if not hasattr(uncertainty_type, "bounded_random_variables"):
            raise UnknownUncertaintyType(
                "The provided uncertainty type must have the "
                "`bounded_random_variables` method."
            )

    def generate_random_numbers(self, uncertainty_type=None, params=None, size=None):
        if not uncertainty_type:
            uncertainty_type = self.uncertainty_type
        if params is None:  # Can't convert array to boolean
            params = self.params
        if not size:
            size = self.size
        return uncertainty_type.bounded_random_variables(
            params, size, self.random, self.maximum_iterations
        )

    def next(self):
        return self.generate_random_numbers()

    def __next__(self):
        return self.next()

    def __iter__(self):
        return self


class MCRandomNumberGenerator(Iterable):
    """
    A Monte Carlo random number generator that operates on a :ref:`hpa`.

    Upon instantiation, the class checks that:

    * Each unique ``uncertainty_type`` is a valid choice in ``uncertainty_choices``
    * That the parameter array for each uncertainty type validates

    The returned class instance can be called directly with ``next``, or can be used as an iterator::

        >>> from stats_arrays import MCRandomNumberGenerator, UncertaintyBase
        >>> params = UncertaintyBase.from_dicts(
        ...     {'loc': 5, 'minimum': 3, 'maximum': 10, 'uncertainty_type': 5},
        ...     {'loc': 1, 'scale': 0.7, 'uncertainty_type': 3}
        ...     )
        >>> mcrng = MCRandomNumberGenerator(params)
        >>> zip(range(2), mcrng)
        [(0, array([ 1.35034874,  5.2705415 ])),
         (1, array([ 5.2705415 ,  1.35034874]))]

    Args:
        * **params** (array): The :ref:`hpa`
        * *maximum_iterations* (int, optional): The number of times to draw samples that fit within the given bounds, if any, before raising ``stats_arrays.MaximumIterationsError``. Default is ``100``.
        * *seed* (int, optional): Seed value for the random number generator. Default is ``None``.

    Returns:
        A class instance

    """

    def __init__(self, params, maximum_iterations=50, seed=None, **kwargs):
        self.params = params.copy()
        self.length = self.params.shape[0]
        self.maximum_iterations = maximum_iterations
        self.choices = uncertainty_choices
        self.random = np.random.RandomState(seed)
        self.verify_params()
        self.ordering = np.argsort(self.params["uncertainty_type"])
        self.reverse_ordering = np.argsort(self.ordering)
        self.params = self.params[self.ordering]
        self.positions = self.get_positions()

    def get_positions(self):
        """Construct dictionary of where each distribution starts and stops in the sorted parameter array"""
        return dict(
            [
                (choice, (self.params["uncertainty_type"] == choice.id).sum())
                for choice in self.choices
            ]
        )

    def verify_params(self):
        """Verify that all uncertainty types are allowed, and parameter validate using distribution class methods"""
        ids = set(np.unique(self.params["uncertainty_type"]))
        extra_ids = ids.difference(set([x.id for x in self.choices]))
        if extra_ids:
            raise ValueError(
                "Uncertainty type id(s) {} are not valid".format(extra_ids)
            )

        for uncertainty_type in self.choices:
            mask = self.params["uncertainty_type"] == uncertainty_type.id
            if mask.sum():
                uncertainty_type.validate(self.params[mask])

    def generate(self, samples=1):
        """Generate random samples.

        If ``samples`` is one, return a one-dimensional array. Otherwise returns a ``num_parameters, samples`` array.
        """
        if samples == 1:
            self.random_data = np.zeros(self.length)
        else:
            self.random_data = np.zeros((self.length, samples))

        offset = 0
        for uncertainty_type in self.choices:
            numparams = self.positions[uncertainty_type]
            if not numparams:
                continue
            random_data = uncertainty_type.bounded_random_variables(
                self.params[offset : numparams + offset],
                samples,
                self.random,
                self.maximum_iterations,
            )
            if samples == 1:
                if len(random_data.shape) == 2:
                    random_data = random_data[:, 0]  # Restore to 1-d
                self.random_data[offset : numparams + offset] = random_data
            else:
                self.random_data[offset : numparams + offset, :] = random_data
            offset += numparams

        self.random_data = self.random_data[self.reverse_ordering]
        return self.random_data

    def next(self):
        """Generate a new vector of random numbers"""
        return self.generate()

    def __next__(self):
        return self.next()

    def __iter__(self):
        return self


class LatinHypercubeRNG(MCRandomNumberGenerator):
    """
    A random number generator that pre-calculates a sample space to draw from.

    .. rubric:: Inputs

    * params : A :ref:`params-array` which gives parameters for distributions (one distribution per row).
    * seed : An integer (or array of integers) to seed the `NumPy random number generator <http://docs.scipy.org/doc/numpy/reference/generated/numpy.random.mtrand.RandomState.html#numpy.random.mtrand.RandomState>`_.
    * samples : An integer number of samples to construct for each distribution.
    """

    def __init__(self, params, seed=None, samples=10, **kwargs):
        self.params = params
        self.length = self.params.shape[0]
        self.row_index = np.arange(self.length)
        self.size = 1
        self.samples = samples
        self.choices = uncertainty_choices
        self.random = np.random.RandomState(seed)
        self.uncertainty_type = None

        self.verify_params()
        self.build_hypercube()

    def build_hypercube(self):
        """Build an array, of shape `self.length` rows by `self.samples` columns, which contains the sample space to be drawn from when doing Latin Hypercubic sampling.

        Each row represents a different data point and distribution. The final sample space is `self.hypercube`. All distributions from `uncertainty_choices` are usable, and bounded distributions are also fine.

        .. rubric:: Builds

        self.hypercube : Numpy array with dimensions `self.length` by `self.samples`."""
        step_size = 1 / (self.samples + 1)
        # Define the beginning points and step sizes - not all the same because
        # some distributions are bounded. Make adjustments to generic values
        # later.
        inputs = np.tile(np.array((step_size, step_size)), self.length).reshape(
            self.length, 2
        )
        for uncertainty_type in self.choices:
            mask = self.params["uncertainty_type"] == uncertainty_type.id
            if not mask.sum():
                continue
            subarray = self.params[mask]
            # Adjust inputs when bounds are present. Easiest to do in three
            # discrete steps. First, when only a lower bound is present.
            only_min_mask = ~np.isnan(subarray["minimum"]) * np.isnan(
                subarray["maximum"]
            )
            if only_min_mask.sum():
                mins = uncertainty_type.cdf(
                    subarray[only_min_mask], subarray[only_min_mask]["minimum"]
                )
                steps = (1 - mins) / (self.samples + 1)
                inputs[mask, :][only_min_mask] = np.hstack((mins + steps, steps))
            # Next, if only a max bound is present
            only_max_mask = np.isnan(subarray["minimum"]) * ~np.isnan(
                subarray["maximum"]
            )
            if only_max_mask.sum():
                maxs = uncertainty_type.cdf(
                    subarray[only_max_mask], subarray[only_max_mask]["maximum"]
                )
                steps = maxs / (self.samples + 1)
                inputs[mask, :][only_max_mask] = np.hstack((steps, steps))
            # Finally, if both min and max bounds are present
            both_mask = ~np.isnan(subarray["minimum"]) * ~np.isnan(subarray["maximum"])
            if both_mask.sum():
                mins = uncertainty_type.cdf(
                    subarray[both_mask], subarray[both_mask]["minimum"]
                )
                maxs = uncertainty_type.cdf(
                    subarray[both_mask], subarray[both_mask]["maximum"]
                )
                steps = (maxs - mins) / (self.samples + 1)
                inputs[mask, :][both_mask] = np.hstack((mins + steps, steps))
        # Percentages is now a list, samples wide, of the percentages covered
        # by the bounded or unbounded distributions.
        self.percentages = inputs[:, 0].reshape(self.length, 1) + np.arange(
            0, self.samples, 1
        ) * inputs[:, 1].reshape(self.length, 1)
        # Transform percentages into a sample space
        self.hypercube = np.zeros((self.length, self.samples))
        for uncertainty_type in self.choices:
            mask = self.params["uncertainty_type"] == uncertainty_type.id
            if not mask.sum():
                continue
            self.hypercube[mask, :] = uncertainty_type.ppf(
                params=self.params[mask], percentages=self.percentages[mask, :]
            )
        self.hypercube[self.params["negative"], :] = (
            -1 * self.hypercube[self.params["negative"], :]
        )

    def next(self):
        """Draw directly from pre-computed sample space."""
        return self.hypercube[
            self.row_index, self.random.randint(self.samples, size=self.length)
        ]



__all__ = (
    "BernoulliUncertainty",
    "BetaUncertainty",
    "BetaPERTUncertainty",
    "BoundedUncertaintyBase",
    "DiscreteUniform",
    "GammaUncertainty",
    "GeneralizedExtremeValueUncertainty",
    "LognormalUncertainty",
    "NormalUncertainty",
    "NoUncertainty",
    "StudentsTUncertainty",
    "TriangularUncertainty",
    "UncertaintyBase",
    "UndefinedUncertainty",
    "UniformUncertainty",
    "WeibullUncertainty",
)

from stats_arrays.distributions.base import BoundedUncertaintyBase, UncertaintyBase
from stats_arrays.distributions.bernoulli import BernoulliUncertainty
from stats_arrays.distributions.beta import BetaUncertainty
from stats_arrays.distributions.discrete_uniform import DiscreteUniform
from stats_arrays.distributions.extreme import GeneralizedExtremeValueUncertainty
from stats_arrays.distributions.gamma import GammaUncertainty
from stats_arrays.distributions.geometric import TriangularUncertainty, UniformUncertainty
from stats_arrays.distributions.lognormal import LognormalUncertainty
from stats_arrays.distributions.normal import NormalUncertainty
from stats_arrays.distributions.student import StudentsTUncertainty
from stats_arrays.distributions.undefined import NoUncertainty, UndefinedUncertainty
from stats_arrays.distributions.weibull import WeibullUncertainty
from stats_arrays.distributions.beta_pert import BetaPERTUncertainty



import numpy as np
import pytest

from stats_arrays.distributions import (
    BoundedUncertaintyBase,
    NormalUncertainty,
    UncertaintyBase,
    UndefinedUncertainty,
)
from stats_arrays.errors import (
    ImproperBoundsError,
    InvalidParamsError,
    UndefinedDistributionError,
    UnreasonableBoundsError,
)


def make_params_array(length=1):
    assert isinstance(length, int)
    params = np.zeros(
        (length,),
        dtype=[
            ("input", "u4"),
            ("output", "u4"),
            ("loc", "f4"),
            ("negative", "b1"),
            ("scale", "f4"),
            ("shape", "f4"),
            ("minimum", "f4"),
            ("maximum", "f4"),
        ],
    )
    params["minimum"] = params["maximum"] = np.nan
    params["scale"] = params["shape"] = np.nan
    return params


def test_uncertainty_base_validate():
    """UncertaintyBase: Mean exists, and bounds are ok if present."""
    params = make_params_array(1)
    params["maximum"] = 2
    params["minimum"] = 2.1
    with pytest.raises(ImproperBoundsError):
        UncertaintyBase.validate(params)


def test_check_2d_inputs():
    params = make_params_array(2)
    params["minimum"] = 0
    params["loc"] = 1
    params["maximum"] = 2
    # Params has 2 rows. The input vector can only have shape (2,) or (2, n)
    with pytest.raises(InvalidParamsError):
        UncertaintyBase.check_2d_inputs(params, np.array((1,)))
    with pytest.raises(InvalidParamsError):
        UncertaintyBase.check_2d_inputs(params, np.array(((1, 2),)))
    with pytest.raises(InvalidParamsError):
        UncertaintyBase.check_2d_inputs(params, np.array(((1, 2), (3, 4), (5, 6))))

    # Test 1-d input
    vector = UncertaintyBase.check_2d_inputs(params, np.array((1, 2)))
    assert np.allclose(vector, np.array(([1], [2])))
    # Test 1-row 2-d input
    vector = UncertaintyBase.check_2d_inputs(params, np.array(((1, 2, 3), (1, 2, 3))))
    assert np.allclose(vector, np.array(((1, 2, 3), (1, 2, 3))))


def test_check_bounds_reasonableness_no_minimum():
    params = make_params_array(1)
    params["maximum"] = -0.3
    params["loc"] = 1
    params["scale"] = 1
    NormalUncertainty.check_bounds_reasonableness(params)


def test_check_bounds_reasonableness():
    params = make_params_array(1)
    params["minimum"] = -100
    params["maximum"] = -0.3
    params["loc"] = 1
    params["scale"] = 1
    with pytest.raises(UnreasonableBoundsError):
        NormalUncertainty.check_bounds_reasonableness(params)


def test_bounded_random_variables():
    params = make_params_array(1)
    params["maximum"] = -0.2  # Only ~ 10 percent of distribution
    params["loc"] = 1
    params["scale"] = 1
    sample = NormalUncertainty.bounded_random_variables(
        params, size=50000, maximum_iterations=1000
    )
    assert (sample > -0.2).sum() == 0
    assert sample.shape == (1, 50000)
    assert np.abs(sample.sum()) > 0


def test_bounded_uncertainty_base_validate():
    """BoundedUncertaintyBase: Make sure legitimate bounds are provided"""
    params = make_params_array(1)
    # Only maximum
    params["maximum"] = 1
    params["minimum"] = np.nan
    with pytest.raises(ImproperBoundsError):
        BoundedUncertaintyBase.validate(params)

    # Only minimum
    params["maximum"] = np.nan
    params["minimum"] = -1
    with pytest.raises(ImproperBoundsError):
        BoundedUncertaintyBase.validate(params)


def test_undefined_uncertainty():
    params = make_params_array(1)
    with pytest.raises(UndefinedDistributionError):
        UndefinedUncertainty.cdf(params, np.random.random(10))
    params = make_params_array(2)
    params["loc"] = 9
    assert np.allclose(
        np.ones((2, 3)) * 9, UndefinedUncertainty.random_variables(params, 3)
    )
    random_percentages = np.random.random(20).reshape(2, 10)
    assert np.allclose(
        np.ones((2, 10)) * 9, UndefinedUncertainty.ppf(params, random_percentages)
    )



from typing import Optional
from numpy import isnan, ndarray, nan, random

from stats_arrays.distributions.beta import BetaUncertainty
from stats_arrays.errors import InvalidParamsError, ImproperBoundsError
from stats_arrays.utils import one_row_params_array


class BetaPERTUncertainty(BetaUncertainty):
    """
    The Beta PERT distribution is a particular form of the Beta distribution which takes these input parameters:

    #. ``minimum``: ``A``, the minimum value
    #. ``loc``: ``B``, the mean value
    #. ``maximum``: ``C``, the maximum value
    #. ``shape``: ``:math:`\\lambda```, the optional shape parameter

    The standard default shape parameter is 4, though according to Wikipedia values from 2 to 3.5 are commonly used. A higher value makes the distribution more concentrated on the mean value.

    Wikipedia: `Beta PERT distribution <https://en.wikipedia.org/wiki/PERT_distribution>`_
    """

    id = 13
    description = "Beta PERT uncertainty"

    @classmethod
    def validate(cls, params):
        if isnan(params["minimum"]).sum():
            raise InvalidParamsError(
                "Real, positive `A` values are required for Beta PERT uncertainties."
            )
        if isnan(params["loc"]).sum():
            raise InvalidParamsError(
                "Real, positive `B` values are required for Beta PERT uncertainties."
            )
        if isnan(params["maximum"]).sum():
            raise InvalidParamsError(
                "Real, positive `C` values are required for Beta PERT uncertainties."
            )
        if (params["minimum"] > params["loc"]).sum() or (
            params["loc"] > params["maximum"]
        ).sum():
            raise ImproperBoundsError("`A <= B <= C` not respected.")
        if (params["minimum"] == params["maximum"]).sum():
            raise ImproperBoundsError("`A` and `C` have the same values.")
        # Check lambda values where they are provided (not NaN)
        provided_lambda = params[~isnan(params["scale"])]
        if (provided_lambda["scale"] <= 0).sum():
            raise InvalidParamsError("Lambda values must be greater than zero.")

    @classmethod
    def _as_beta(cls, params: ndarray, default_lambda: float = 4.0) -> ndarray:
        """Calculate α and β values for Beta distribution from PERT A/B/C inputs."""
        beta = params.copy()
        # Use provided lambda values or default
        lambda_values = params["scale"].copy()
        missing_lambda_mask = isnan(lambda_values)
        lambda_values[missing_lambda_mask] = default_lambda

        a, b, c = params["minimum"], params["loc"], params["maximum"]
        # Set Beta distribution parameters
        beta["scale"] = nan
        beta["loc"] = 1 + lambda_values * ((b - a) / (c - a))  # alpha
        beta["shape"] = 1 + lambda_values * ((c - b) / (c - a))  # beta
        # Set minimum and maximum for scaling from [0, 1] to [A, C]
        beta["minimum"] = a
        beta["maximum"] = c
        return beta

    @classmethod
    def random_variables(
        cls,
        params: ndarray,
        size: int,
        seeded_random: Optional[random.RandomState] = None,
        transform: bool = False,
        default_lambda: float = 4.0,
    ):
        return BetaUncertainty.random_variables(
            params=cls._as_beta(params=params, default_lambda=default_lambda),
            size=size,
            seeded_random=seeded_random,
            transform=transform,
        )

    @classmethod
    def cdf(cls, params: ndarray, vector: ndarray, default_lambda: float = 4.0):
        return BetaUncertainty.cdf(
            params=cls._as_beta(params=params, default_lambda=default_lambda),
            vector=vector,
        )

    @classmethod
    def ppf(cls, params: ndarray, percentages: ndarray, default_lambda: float = 4.0):
        return BetaUncertainty.ppf(
            params=cls._as_beta(params=params, default_lambda=default_lambda),
            percentages=percentages,
        )

    @classmethod
    @one_row_params_array
    def statistics(cls, params: ndarray, default_lambda: float = 4.0):
        return BetaUncertainty.statistics(
            params=cls._as_beta(params=params, default_lambda=default_lambda)
        )

    @classmethod
    @one_row_params_array
    def pdf(
        cls, params: ndarray, xs: Optional[ndarray] = None, default_lambda: float = 4.0
    ):
        return BetaUncertainty.pdf(
            params=cls._as_beta(params=params, default_lambda=default_lambda), xs=xs
        )



from __future__ import division

import numpy as np

from stats_arrays.errors import InvalidParamsError
from stats_arrays.distributions.base import UncertaintyBase


class StudentsTUncertainty(UncertaintyBase):
    """
    The Student's T uncertainty distribution probability density function is a function of :math:`\\nu`, the degrees of freedom:

    .. math:: f(x; \\nu) = \\frac{\\Gamma(\\frac{\\nu+1}{2})} {\\sqrt{\\nu\\pi},\\Gamma(\\frac{\\nu}{2})} \\left(1+\\frac{x^2}{\\nu} \\right)^{-\\frac{\\nu+1}{2}}

    A non-standardized distribution, with a location and scale parameter, is also possible, through the transformation:

    .. math:: X = \\mu + \\sigma f

    In our implementation, the location parameter :math:`\\mu` is ``location``, the scale parameter :math:`\\sigma` is ``scale``, and :math:`\\nu` (the degrees of freedom)  is ``shape``.

    See http://en.wikipedia.org/wiki/Student%27s_t-distribution
    """

    id = 12
    description = "Student's T uncertainty"

    @classmethod
    def validate(cls, params):
        if (params["shape"] <= 0).sum() or np.isnan(params["shape"]).sum():
            raise InvalidParamsError(
                "Positive ``nu`` (degrees of freedom) values are required for"
                " Student's T."
            )
        if (params["scale"] <= 0).sum():
            raise InvalidParamsError(
                "Scale values, if specified, must be greater than zero for"
                " Student's T."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None, **kwargs):
        if seeded_random is None:
            seeded_random = np.random
        scale = params["scale"].copy()
        scale[np.isnan(scale)] = 1
        location = params["loc"].copy()
        location[np.isnan(location)] = 0
        data = (
            location
            + scale
            * seeded_random.standard_t(params["shape"], size=(size, params.shape[0])).T
        )
        return data



import numpy as np
import pytest

from stats_arrays.distributions import (
    NormalUncertainty,
    TriangularUncertainty,
    UncertaintyBase,
)
from stats_arrays.errors import ImproperBoundsError, UnknownUncertaintyType
from stats_arrays.random import RandomNumberGenerator as RNG


def test_invalid_uncertainty_type():
    with pytest.raises(UnknownUncertaintyType):
        RNG(object, UncertaintyBase.from_dicts({}))


def test_uncertainty_not_subclass():
    class Foo(object):

        @classmethod
        def bounded_random_variables():
            pass

        @classmethod
        def validate(self, *args, **kwargs):
            pass

    RNG(Foo, UncertaintyBase.from_dicts({}))


def test_method_call():
    rng = RNG(NormalUncertainty, UncertaintyBase.from_dicts({"loc": 0, "scale": 1}))
    rng.next()
    rng.generate_random_numbers()


def test_as_iterator():
    counter = 0
    rng = RNG(NormalUncertainty, UncertaintyBase.from_dicts({"loc": 0, "scale": 1}))
    for x in rng:
        counter += 1
        if counter >= 10:
            break


def test_seed():
    data = []
    for x in range(2):
        rng = RNG(
            NormalUncertainty,
            UncertaintyBase.from_dicts({"loc": 0, "scale": 1}),
            seed=111,
        )
        data.append(rng.next())
    assert np.allclose(*data)


def test_output_dimensions():
    rng = RNG(
        NormalUncertainty,
        UncertaintyBase.from_dicts({"loc": 0, "scale": 1}, {"loc": 1, "scale": 2}),
        size=10,
    )
    assert rng.next().shape == (2, 10)


def test_validation():
    with pytest.raises(ImproperBoundsError):
        RNG(
            TriangularUncertainty,
            UncertaintyBase.from_dicts(
                {"loc": -0.000000001, "minimum": 0, "maximum": 1}
            ),
        )

    # No error
    RNG(
        TriangularUncertainty,
        UncertaintyBase.from_dicts({"loc": 0.5, "minimum": 0, "maximum": 1}),
    )
    RNG(
        TriangularUncertainty,
        UncertaintyBase.from_dicts({"loc": 0.0, "minimum": 0, "maximum": 1}),
    )
    RNG(
        TriangularUncertainty,
        UncertaintyBase.from_dicts({"loc": 1.0, "minimum": 0, "maximum": 1}),
    )



import numpy as np

from stats_arrays.distributions import UncertaintyBase
from stats_arrays.utils import construct_params_array


def sa_allclose(a, b):
    """allclose for structured arrays"""
    for name in a.dtype.names:
        nan_mask_a = np.isnan(a[name])
        nan_mask_b = np.isnan(b[name])
        assert np.allclose(nan_mask_a, nan_mask_b)
        assert np.allclose(a[name][~nan_mask_a], b[name][~nan_mask_b])


def get_right_1d_array():
    params = construct_params_array(include_type=True)
    params["loc"] = 1
    params["maximum"] = 2
    params["minimum"] = 0
    params["scale"] = 3
    return params


def get_right_2d_array():
    params = construct_params_array(2, include_type=True)
    params["loc"] = (1, 2)
    params["scale"] = (3, 4)
    params["maximum"] = 10
    params["minimum"] = 0
    return params


def test_1d_dict():
    values = {
        "loc": 1.0,
        "scale": 3.0,
        "maximum": 2.0,
        "minimum": 0.0,
        "negative": False,
    }
    sa_allclose(
        UncertaintyBase.from_dicts(values),
        get_right_1d_array(),
    )


def test_1d_tuple():
    values = (1, 3, np.nan, 0, 2, False, 0)
    sa_allclose(
        UncertaintyBase.from_tuples(values),
        get_right_1d_array(),
    )


def test_2d_dict():
    values = (
        {
            "loc": 1.0,
            "scale": 3.0,
            "maximum": 10.0,
            "minimum": 0.0,
            "negative": False,
        },
        {
            "loc": 2.0,
            "scale": 4.0,
            "maximum": 10.0,
            "minimum": 0.0,
            "negative": False,
        },
    )
    sa_allclose(
        UncertaintyBase.from_dicts(*values),
        get_right_2d_array(),
    )


def test_2d_tuple():
    values = (
        (1, 3, np.nan, 0, 10, False, 0),
        (2, 4, np.nan, 0, 10, False, 0),
    )
    sa_allclose(
        UncertaintyBase.from_tuples(*values),
        get_right_2d_array(),
    )


def test_hetergeneous_from_dicts():
    dicts = [
        {
            "uncertainty type": 4,
            "minimum": 2,
            "maximum": 5,
            "amount": 3,
            "loc": 3,
        },
        {
            "uncertainty_type": 2,
            "loc": 3,
            "scale": 0.2,
        },
    ]
    answer = UncertaintyBase.from_dicts(*dicts)
    assert np.allclose(answer["uncertainty_type"], np.array((4, 2)))


def test_without_underscore():
    dicts = [
        {
            "uncertainty type": 4,
            "minimum": 2,
            "maximum": 5,
            "amount": 3,
            "loc": 3,
        }
    ]
    answer = UncertaintyBase.from_dicts(*dicts)
    assert np.allclose(answer["uncertainty_type"][0], 4)



from __future__ import division

import numpy as np
from scipy import stats

from stats_arrays.errors import InvalidParamsError
from stats_arrays.utils import one_row_params_array
from stats_arrays.distributions.base import UncertaintyBase


class NormalUncertainty(UncertaintyBase):
    id = 3
    description = "Normal uncertainty"

    @classmethod
    def validate(cls, params):
        if np.isnan(params["scale"]).sum() or (params["scale"] <= 0).sum():
            raise InvalidParamsError(
                "Real, positive scale (sigma) values are required"
                " for normal uncertainties."
            )
        if np.isnan(params["loc"]).sum():
            raise InvalidParamsError(
                "Real loc (mu) values are required for normal uncertainties."
            )

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        return seeded_random.normal(
            params["loc"], params["scale"], size=(size, params.shape[0])
        ).T

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = np.zeros(vector.shape)
        for row in range(params.shape[0]):
            results[row, :] = stats.norm.cdf(
                vector[row, :], loc=params["loc"][row], scale=params["scale"][row]
            )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        results = np.zeros(percentages.shape)
        for row in range(percentages.shape[0]):
            results[row, :] = stats.norm.ppf(
                percentages[row, :], loc=params["loc"][row], scale=params["scale"][row]
            )
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        return {
            "mean": float(params["loc"].flat[0]),
            "mode": float(params["loc"].flat[0]),
            "median": float(params["loc"].flat[0]),
            "lower": float((params["loc"] - 2 * params["scale"]).flat[0]),
            "upper": float((params["loc"] + 2 * params["scale"]).flat[0]),
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        if xs is None:
            if np.isnan(params["minimum"]):
                lower = (
                    params["loc"]
                    - params["scale"] * cls.standard_deviations_in_default_range
                )
            else:
                lower = params["minimum"]
            if np.isnan(params["maximum"]):
                upper = (
                    params["loc"]
                    + params["scale"] * cls.standard_deviations_in_default_range
                )
            else:
                upper = params["maximum"]
            xs = np.arange(
                lower.flat[0],
                upper.flat[0],
                (upper - lower).flat[0] / cls.default_number_points_in_pdf,
            )

        ys = stats.norm.pdf(xs, params["loc"], params["scale"])
        return xs, ys.reshape(ys.shape[1])



import pytest

from stats_arrays import (
    MCRandomNumberGenerator,
    NormalUncertainty,
    TriangularUncertainty,
    UncertaintyBase,
)


@pytest.fixture()
def mv():
    return UncertaintyBase.from_dicts(
        {"loc": 2, "scale": 0.5, "uncertainty_type": NormalUncertainty.id},
        {
            "loc": 1.5,
            "minimum": 0,
            "maximum": 10,
            "uncertainty_type": TriangularUncertainty.id,
        },
    )


@pytest.fixture()
def mrng(mv):
    return MCRandomNumberGenerator(mv)


def test_generate(mrng):
    assert mrng.generate(10).shape == (2, 10)
    assert mrng.generate(1).shape == (2,)
    assert mrng.generate().shape == (2,)
    assert mrng.generate(10).shape == (2, 10)


def test_next(mrng):
    assert mrng.next().shape == (2,)
    assert next(mrng).shape == (2,)


def test_iterable(mrng):
    for x, y in zip(mrng, range(10)):
        assert x.shape == (2,)



__version__ = "0.8"

__all__ = (
    "BernoulliUncertainty",
    "BetaUncertainty",
    "BetaPERTUncertainty",
    "BoundedUncertaintyBase",
    "DiscreteUniform",
    "GammaUncertainty",
    "GeneralizedExtremeValueUncertainty",
    "ImproperBoundsError",
    "InvalidParamsError",
    "LatinHypercubeRNG",
    "LognormalUncertainty",
    "MaximumIterationsError",
    "MCRandomNumberGenerator",
    "MultipleRowParamsArrayError",
    "NormalUncertainty",
    "NoUncertainty",
    "RandomNumberGenerator",
    "StatsArraysError",
    "StudentsTUncertainty",
    "TriangularUncertainty",
    "uncertainty_choices",
    "UncertaintyBase",
    "UndefinedDistributionError",
    "UndefinedUncertainty",
    "UniformUncertainty",
    "UnknownUncertaintyType",
    "UnreasonableBoundsError",
    "WeibullUncertainty",
)

from stats_arrays.distributions import (
    BernoulliUncertainty,
    BetaUncertainty,
    BetaPERTUncertainty,
    BoundedUncertaintyBase,
    DiscreteUniform,
    GammaUncertainty,
    GeneralizedExtremeValueUncertainty,
    LognormalUncertainty,
    NormalUncertainty,
    NoUncertainty,
    StudentsTUncertainty,
    TriangularUncertainty,
    UncertaintyBase,
    UndefinedUncertainty,
    UniformUncertainty,
    WeibullUncertainty,
)
from stats_arrays.errors import (
    ImproperBoundsError,
    InvalidParamsError,
    MaximumIterationsError,
    MultipleRowParamsArrayError,
    StatsArraysError,
    UndefinedDistributionError,
    UnknownUncertaintyType,
    UnreasonableBoundsError,
)
from stats_arrays.random import LatinHypercubeRNG, MCRandomNumberGenerator, RandomNumberGenerator
from stats_arrays.uncertainty_choices import uncertainty_choices



from __future__ import division

from numpy import isnan, linspace, random, zeros, newaxis
from scipy import stats

from stats_arrays.errors import InvalidParamsError, ImproperBoundsError
from stats_arrays.utils import one_row_params_array
from stats_arrays.distributions.base import UncertaintyBase


class BetaUncertainty(UncertaintyBase):
    """
    The Beta distribution has the probability distribution function:

    .. math:: f(x; \\alpha, \\beta) = \\frac{1}{B(\\alpha, \\beta)} x^{\\alpha - 1}(1 - x)^{\\beta - 1},

    where the normalisation, *B*, is the beta function:

    .. math:: B(\\alpha, \\beta) = \\int_0^1 t^{\\alpha - 1}(1 - t)^{\\beta - 1} dt

    The :math:`\\alpha` parameter is ``loc``, and :math:`\\beta` is ``shape``. By default, the Beta distribution is defined from 0 to 1; the lower and upper bounds can be rescaled with the ``minimum`` and ``maximum`` parameters.

    Wikipedia: `Beta distribution <http://en.wikipedia.org/wiki/Beta_distribution>`_
    """

    id = 10
    description = "Beta uncertainty"

    @classmethod
    def validate(cls, params):
        if (params["loc"] > 0).sum() != params.shape[0]:
            raise InvalidParamsError(
                "Real, positive alpha values are" + " required for Beta uncertainties."
            )
        if (params["shape"] > 0).sum() != params.shape[0]:
            raise InvalidParamsError(
                "Real, positive beta values are" + " required for Beta uncertainties."
            )
        if (params["minimum"] >= params["maximum"]).sum() or (
            params["maximum"] <= params["minimum"]
        ).sum():
            raise ImproperBoundsError("Min/max inconsistency.")

    @classmethod
    def _rescale(cls, params, results):
        """Rescale results from [0, 1] to [minimum, maximum] range.

        The correct scaling formula is: results = results * scale + minimum
        where scale = maximum - minimum.
        """
        minimum, scale = cls._minimum_scale(params)

        # Handle broadcasting for multiple rows
        if results.ndim == 2 and minimum.ndim == 1:
            # results shape: (n_rows, n_samples), minimum/scale shape: (n_rows,)
            results = results * scale[:, newaxis] + minimum[:, newaxis]
        else:
            # Single row case or matching dimensions
            results = results * scale + minimum

        return results

    @classmethod
    def _minimum_scale(cls, params):
        minimum = params["minimum"].copy()
        minimum[isnan(minimum)] = 0
        scale = params["maximum"].copy()
        scale[isnan(scale)] = 1
        scale -= minimum
        return minimum, scale

    @classmethod
    def random_variables(cls, params, size, seeded_random=None, transform=False):
        if not seeded_random:
            seeded_random = random
        return cls._rescale(
            params,
            seeded_random.beta(
                params["loc"], params["shape"], size=(size, params.shape[0])
            ).T,
        )

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = zeros(vector.shape)
        loc, scale = cls._minimum_scale(params)
        for row in range(params.shape[0]):
            results[row, :] = stats.beta.cdf(
                vector[row, :],
                params["loc"][row],
                params["shape"][row],
                loc=loc[row],
                scale=scale[row],
            )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        results = zeros(percentages.shape)
        loc, scale = cls._minimum_scale(params)
        for row in range(percentages.shape[0]):
            results[row, :] = stats.beta.ppf(
                percentages[row, :],
                params["loc"][row],
                params["shape"][row],
                loc=loc[row],
                scale=scale[row],
            )
        return results

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        alpha = float(params["loc"][0][0])
        beta = float(params["shape"][0][0])

        minimum, scale = cls._minimum_scale(params)
        minimum = float(minimum[0][0])
        scale = float(scale[0][0])

        if alpha <= 1 or beta <= 1:
            mode = "Undefined"
        else:
            mode = ((alpha - 1) / (alpha + beta - 2)) * scale + minimum
        return {
            "mean": (alpha / (alpha + beta)) * scale + minimum,
            "mode": mode,
            "median": "Not Implemented",
            "lower": "Not Implemented",
            "upper": "Not Implemented",
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        loc, scale = cls._minimum_scale(params)
        loc = float(loc[0][0])
        scale = float(scale[0][0])

        if xs is None:
            xs = linspace(loc, loc + scale, cls.default_number_points_in_pdf)
        ys = stats.beta.pdf(xs, params["loc"], params["shape"], loc=loc, scale=scale)
        return xs, ys.reshape(ys.shape[1])



from __future__ import division

from numpy import repeat, tile

from stats_arrays.errors import UndefinedDistributionError
from stats_arrays.distributions.base import UncertaintyBase


class UndefinedUncertainty(UncertaintyBase):
    """Undefined or unknown uncertainty"""

    id = 0
    description = "Undefined or unknown uncertainty"

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        return repeat(params["loc"], size).reshape((params.shape[0], size))

    @classmethod
    def cdf(cls, params, vector):
        raise UndefinedDistributionError(
            "Can't calculate percentages for an undefined distribution."
        )

    @classmethod
    def ppf(cls, params, percentages):
        return tile(params["loc"].reshape((params.shape[0], 1)), percentages.shape[1])


class NoUncertainty(UndefinedUncertainty):
    id = 1
    description = "No uncertainty"



# -*- coding: utf-8 -*
import numpy as np
from scipy import stats

from stats_arrays.errors import ImproperBoundsError, InvalidParamsError
from stats_arrays.utils import one_row_params_array
from stats_arrays.distributions.base import UncertaintyBase


class DiscreteUniform(UncertaintyBase):
    """
    The discrete uniform distribution includes all integer values from the ``minimum`` up to, but excluding the ``maximum``.

    See https://en.wikipedia.org/wiki/Uniform_distribution_(discrete).
    """

    id = 7
    description = "Discrete uniform uncertainty"

    @classmethod
    def validate(cls, params):
        # No mean value
        if np.isnan(params["maximum"]).sum():
            raise InvalidParamsError("Maximum values must always be defined.")
        # Minimum <= Maximum
        if (params["minimum"] >= params["maximum"]).sum():
            raise ImproperBoundsError

    @classmethod
    def fix_nan_minimum(cls, params):
        mask = np.isnan(params["minimum"])
        params["minimum"][mask] = 0
        return params

    @classmethod
    def random_variables(cls, params, size, seeded_random=None):
        if not seeded_random:
            seeded_random = np.random
        params = cls.fix_nan_minimum(params)
        # randint has different behaviour than e.g. uniform. We can't pass in
        # arrays, but have to process them line by line.
        return np.vstack(
            [
                seeded_random.randint(
                    params["minimum"][i],  # Minimum (low)
                    params["maximum"][i],  # Maximum (high)
                    size=size,
                )
                for i in range(params.shape[0])
            ]
        )

    @classmethod
    def cdf(cls, params, vector):
        vector = cls.check_2d_inputs(params, vector)
        results = np.zeros(vector.shape)
        params = cls.fix_nan_minimum(params)
        for row in range(params.shape[0]):
            results[row, :] = stats.randint.cdf(
                vector[row, :],
                loc=params[row]["minimum"],
                scale=params[row]["maximum"] - params[row]["minimum"],
            )
        return results

    @classmethod
    def ppf(cls, params, percentages):
        percentages = cls.check_2d_inputs(params, percentages)
        params = cls.fix_nan_minimum(params)
        scale = (params["maximum"] - params["minimum"]).reshape(params.shape[0], 1)
        return percentages * scale + params["minimum"].reshape(params.shape[0], 1)

    @classmethod
    @one_row_params_array
    def statistics(cls, params):
        params = cls.fix_nan_minimum(params)
        mean = (params["maximum"] + params["minimum"]) / 2
        return {
            "mean": mean,
            "mode": "Undefined",
            "median": int(mean.round(0)),
            "lower": params["minimum"],
            "upper": params["maximum"],
        }

    @classmethod
    @one_row_params_array
    def pdf(cls, params, xs=None):
        params = cls.fix_nan_minimum(params)
        if xs is None:
            xs = (params["minimum"], params["maximum"])
        percentage = 1 / (params["maximum"] - params["minimum"])
        ys = np.array([float(percentage) for x in xs])
        return np.array([float(x) for x in xs]), ys
